{{
=====================================================================
SimCache.spin - defines a device driver for SpinSim memory
=====================================================================

    Change History
    --------------
}}

{{
    Copyright (c) 2010-2011 by John Steven Denson (jazzed)
    TERMS OF USE: MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
}}


{{
=====================================================================
Constants to define the simulator functions:
}}
con
    SYS_COMMAND = $12340000
    SYS_LOCKNUM = $12340004
    SYS_PARM    = $12340004
    SYS_EXTMEM_READ   = 17
    SYS_EXTMEM_WRITE  = 18
    SYS_EXTMEM_ALLOC  = 19

    SYS_MEMSIZE    = 64*1024

obj xm : "extmem"
    s  : "conio"

con
{{
=====================================================================
These constants define the driver commands.
Currently, the driver doesn't really care if the command
is a read or write. Use these for future compatibility.
}}
con
    CMD_MASK    = $3
    READ_CMD    = $3
    WRITE_CMD   = $2
    DIRTY_BIT   = $1

{{
=====================================================================
Driver mailbox interface - these longs must stay in this order.
}}
var long    mbcmd, mbdata, mbtags, mbcache

' cog store
var long    cog

con
{{
=====================================================================
Cache is organized as a table.
The tagline table contains pointers to cache entries.
Each tagline has the following definition:

    TAGLINE:    xxxx_xxxi
    such that:  x=match address i=ignored bits

The actual number of ignored bits depends on the cache line length.
}}

con
    CACHESIZE   =  $800>>2          ' power of 2 cache size ... in number of LONGs

con
    TAGCOUNT    = CACHESIZE/LINELEN ' cache with linelen bytes
    LINESHFT    = 5                 ' this must match the size of the buffer operations

    LINELEN     = 1<<LINESHFT       ' CACHELINE SIZE

    TAGMASK     = $003F_FFFF        ' $3FFFFF*512 = 2GB $3FFFFF*32 = 128MB
    DIRTYSHFT   = 31                ' shift by N to get dirty bit   test tag,DIRTYMASK wc
    DIRTYMASK   = 1 << DIRTYSHFT    ' dirty bit mask                muxc tag,DIRTYSHFT

{
{{
=====================================================================
cache
}}
dat ' cache
'tagram          long 0 [TAGCOUNT]
cache           long 0 [CACHESIZE-$1e0]     ' reclaim PASM code for cache

{{
=====================================================================
start(mbptr)
mbptr = start address of command interface
Example:
var long command, data
pub main
  start(@command)
  repeat
}}
pub start(mbptr)
    cache  := $AAAAAAAA
    mbcmd  := mbptr
    mbdata := mbptr+4
    mbtags := @tagram           ' tags are in-cog now - recycle for simulator
    mbcache:= @cache
    long[mbcmd]~~               ' set now so we know when startup is done

    cog := cognew(@pasm,@mbcmd)+1
    repeat while long[mbcmd]    ' wait for startup to finish
    return cog                  ' start failed if return is 0

    cognew(spincache, &stack)

{{
=====================================================================
read cache buffer
madr - cache address to read
returns value read
}}
pub readCache(madr)
    return cache[madr]

{{
=====================================================================
zero cache buffer
clears contents of the cache
}}
pub zeroCache
    bytefill(@cache,0,CACHESIZE)
}

{{
=====================================================================
startx(mbptr,cacheptr)
mbptr = start address of command interface
Example:
obj sd "SdramCache"
var long command, data
var long cache[sd#CACHESIZE]
pub main
  start(@command,@cache)
  repeat
}}
pub startx(mbptr,cptr)
    mbcmd  := mbptr
    mbdata := mbptr+4
    mbcache:= cptr              ' get cache pointer from user
    'initialize simulator memory
    ifnot xm.ExtMemAlloc(0,SYS_MEMSIZE)
        s.str(string($d,$a,"Can't allocate memory "))
        repeat
    long[mbcmd]~~               ' set now so we know when startup is done
    cog := cognew(@pasm,@mbcmd)+1
    repeat while long[mbcmd]    ' wait for startup to finish
    return cog                  ' start failed if return is 0

{{
=====================================================================
stop
}}
pub stop
    if cog                      ' if cog is zero, don't stop
        cogstop(cog~ - 1)       ' stop cog and set cog var to zero

{{
=====================================================================
dataptr - returns address of cache data buffer
}}
pub dataptr
    return long[mbdata]

{{
=====================================================================
mboxptr  - returns address of cache command mailbox
}}
pub mboxptr
    return long[mbcmd]


{{
=====================================================================
}}
pub memsize
    return SYS_MEMSIZE

{{
=====================================================================
read a long
madr - long address to read
returns value read
}}
pub readLong(madr)
    long[mbcmd] := madr | READ_CMD
    repeat while long[mbcmd]
    madr &= LINELEN-1
    return long[long[mbdata]+madr]

{{
=====================================================================
write a long
madr - long address to write
val  - value to write
}}
pub writeLong(madr,val)
    long[mbcmd] := madr | WRITE_CMD
    repeat while long[mbcmd]
    madr &= LINELEN-1
    ' sending write command says this data will be written on next cache flush
    long[long[mbdata]+madr] := val

{{
=====================================================================
read a word
madr - word address to read
returns value read
}}
pub readWord(madr)
    long[mbcmd] := madr | READ_CMD
    repeat while long[mbcmd]
    madr &= LINELEN-1
    return word[long[mbdata]+madr]

{{
=====================================================================
write a word
madr - word address to write
val  - value to write
}}
pub writeWord(madr,val)
    long[mbcmd] := madr | WRITE_CMD
    repeat while long[mbcmd]
    madr &= LINELEN-1
    ' sending write command says this data will be written on next cache flush
    word[long[mbdata]+madr] := val

{{
=====================================================================
read a byte
madr - byte address to read
returns value read
}}
pub readByte(madr)
    long[mbcmd] := madr | READ_CMD
    repeat while long[mbcmd]
    madr &= LINELEN-1
    return byte[long[mbdata]+madr]

{{
=====================================================================
write a byte
madr - byte address to write
val  - value to write
}}
pub writeByte(madr,val)
    long[mbcmd] := madr | WRITE_CMD
    repeat while long[mbcmd]
    madr &= LINELEN-1
    ' sending write command says this data will be written on next cache flush
    byte[long[mbdata]+madr] := val

{{
=====================================================================
read buffer
}}
pub readBuffer(madr,datp) | n
    long[mbcmd] := madr | READ_CMD
    repeat while long[mbcmd]
    longmove(datp,long[mbdata],LINELEN>>2)

{{
=====================================================================
write buffer
}}
pub writeBuffer(madr,datp) | n
    long[mbcmd] := madr | WRITE_CMD
    repeat while long[mbcmd]
    ' sending write command says this buffer will be written on next flush
    longmove(long[mbdata],datp,LINELEN>>2)

{{
=====================================================================
read cache tag line
madr - tag line address entry to read
returns value read
}}
pub readTags(madr)
    ' tagram is in cog now, so we're faking the address
    return (madr/constant(LINELEN>>2))<<LINESHFT
'   return tagram[madr/constant(LINELEN>>2)]

{{
=====================================================================
PASM Cache Driver
=====================================================================
}}
dat                 org 0
pasm
long    0 [8]                               ' debugger stub space - or beginning of tag RAM

'--------------------------------------------------------------------
' startup and some variables
'
                        jmp #setup              ' 0
long    0 [64-$]                            ' tagram space

'====================================================================
' mailbox command spinner loop
' user will get a block pointer back and access the block with a mod index such as:
'
'                       mov     index,addr          ' user must do something like this for return pointer array index
'                       and     index,#(LINELEN-1)  ' index can be only up to LINELEN-1
'
cmdloop
                        rdlong  addr,cmdptr    wz   ' get command/address ... command is in bit 0..1
    if_z                jmp     #cmdloop            ' if zero, do nothing
                        mov     clineptr,addr       ' get the cache line pointer offset from address

                        shr     addr,#LINESHFT wc   ' test for read/write into carry
                        movs    readtag,addr        ' (addr / LINESIZE) mod TAGCOUNT is tag / cache line offset
                        andn    readtag,#$200-TAGCOUNT ' user sends full address. we only load blocks
'                       add     readtag,#8          ' for debugger only
                        muxnc   dirty,_TAG_DIRTY    ' save new tag dirty state - fill pipe slot
readtag                 mov     tag,0-0             ' get tag line contents

                        cmpsub  tag,_TAG_DIRTY wc   ' get the old dirty tag
                        cmp     addr,tag wz
                        muxc    tag,_TAG_DIRTY      ' restore the old dirty tag

    if_e                wrlong  zero,cmdptr         ' if match let user know we're done early ...
                                                     ' we rely on the fact that the user is looking for 0 cmdptr
                                                    ' and must use another HUB access to get data.
                                                    ' prolems may come if the user's cog is lower than this cog number

                        and     clineptr,_CLP_MOD   ' user sends full address. we only load blocks
                        add     clineptr,cacheptr   ' get cache line
                        wrlong  clineptr,datptr     ' send cache buffer to user - data may change before we're done

    if_ne               call    #flush              ' if bad tag, flush - cache code never changes Z flag

writeTag                movd    tagit,readtag
                        or      tag,dirty           ' save new tag dirty state
tagit                   mov     0-0,tag

                        jmp     #cmdloop            ' 18*4 cycles on cache hit, but 16*4 for delivery.

'--------------------------------------------------------------------
' flush method saves dirty page and loads page from backstore
'
flush                   ' if dirty bit set, write old buffer before read

    if_nc               jmp     #flush_rdonly
flush_write             mov     saveaddr,addr       ' save addr for read
                        mov     addr,tag            ' get cacheline tag addr
                        shl     addr,#LINESHFT      ' physical address to write cacheline

                        call    #ExtMemWrite

                        mov     addr,saveaddr       ' restore addr for read
flush_rdonly            mov     tag,addr            ' move address clears dirty/count
                        shl     addr,#LINESHFT      ' physical address to write cacheline

                        call    #ExtMemRead

                        wrlong  zero,cmdptr         ' let user know we're ready after flush ...

flush_ret               ret


{{
=====================================================================
simulator external memory functions
}}
ExtMemRead              mov     ExtMemCommand, #SYS_EXTMEM_READ
                        jmp     #SystemCall

ExtMemWrite             mov     ExtMemCommand, #SYS_EXTMEM_WRITE

                        'Set up the parameter list in hub RAM
SystemCall              mov     temp, par
                        wrlong  addr, temp
                        add     temp, #4
                        wrlong  clineptr, temp
                        add     temp, #4
                        wrlong  NumByte, temp

                        'Wait for lock not set
:loop1                  lockset locknum                       wc
        if_c            jmp     #:loop1

                        'Write the address of the parmeter list
                        wrlong  par, SysParm

                        'Write the external memory command
                        wrword  ExtMemCommand, SysCommand

                        'Wait for the command to be completed
:loop2                  rdword  ExtMemCommand, SysCommand
                        tjnz    ExtMemCommand, #:loop2

                        'Clear the lock
                        lockclr locknum
SystemCall_ret
ExtMemRead_ret
ExtMemWrite_ret         ret

HubAddr                 long    0
ExtMemAddr              long    0
NumByte                 long    LINELEN
ExtMemCommand           long    0
locknum                 long    0
SysCommand              long    SYS_COMMAND
SysLocknum              long    SYS_LOCKNUM
SysParm                 long    SYS_PARM

'--------------------------------------------------------------------
' setup hardware and mailbox interface
'
setup

' get mailbox pointers
temp                    mov     tmp,par
                        rdlong  cmdptr,tmp
                        add     tmp,#4
                        rdlong  datptr,tmp
                        add     tmp,#4
                        rdlong  tagptr,tmp
                        add     tmp,#4
                        rdlong  cacheptr,tmp
                        wrlong  zero,cmdptr
                        jmp     #cmdloop

_TAG_DIRTY      long DIRTYMASK              ' use to mark pages for write-back
_CLP_MOD        long (TAGCOUNT-1)<<LINESHFT ' use modulo for finding tag address

cmdptr          long 0                      ' cache command pointer
datptr          long 0                      ' cache data pointer
tagptr          long 0                      ' tag pointer - recycle for simulator
cacheptr        long 0                      ' cache pointer

bank            long 0
ndx             long 0
tmp             long 0

bufp            long 0
saveaddr        long 0
index           long 0
clineptr        long 0

tags            long 0
tag             long 0
addr            long 0

dirty           long 0
zero            long 0

'====================================================================

_LINELEN        long LINELEN                    ' greatest address for cache
_MAX_ABIT       long $0100_0000

_TAG_MASK       long TAGMASK                    ' greatest address for cache
_TAG_MOD        long (TAGCOUNT-1)<<2            ' use modulo for finding tag address
_TAG_OFF        long TAGCOUNT<<2                ' offset for tags
_TAG_LINE_MASK  long (TAGCOUNT-1)<<LINESHFT     ' use for modulo
_TAG_MSB        long $8000_0000
_TAG_INC        long $0080_0000
_TAG_CNT        long $FF80_0000
_TAG_CMD        long $C000_0000                 ' used for decoding command
_CLP_OFF        long TAGCOUNT<<LINESHFT         ' offset of second cache space

'====================================================================

fillpasm        long 0 [$1ef-$]
endpasm         fit $1ef
